import functools

from mainline.exceptions import UnprovidableError
from mainline.scope import ScopeRegistry, NoneScope
_sentinel = object()


class IProvider(object):

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return self.provide(*args, **kwargs)

    def provide(self, *args, **kwargs):
        raise NotImplementedError

    def has_instance(self):
        raise NotImplementedError

    def set_instance(self, instance):
        raise NotImplementedError

    @property
    def providable(self):
        raise NotImplementedError


class IFactoryProvider(IProvider):

    def __init__(self, factory=None):
        self.set_factory(factory)

    def set_factory(self, factory):
        self.factory = factory

    def has_factory(self):
        return bool(self.factory)

    def provide(self, *args, **kwargs):
        if not self.providable:
            raise UnprovidableError(self)
        return self.factory(*args, **kwargs)

    def has_instance(self):
        return False

    @property
    def providable(self):
        return self.has_instance() or self.has_factory()


class Provider(IFactoryProvider):
    scopes = ScopeRegistry()

    def __init__(self, factory, scope=NoneScope, key=''):
        self.key = key
        self.scope = self.scopes.resolve(scope)
        super(Provider, self).__init__(factory)

    def __repr__(self):
        return '<%s factory=%s scope=%s>' % (self.__class__.__name__, self.factory, self.scope)

    def provide(self, *args, **kwargs):
        if self.has_instance():
            return self.scope[self.key]
        instance = super(Provider, self).provide(*args, **kwargs)
        self.set_instance(instance)
        return instance

    def has_instance(self):
        return self.key in self.scope

    def set_instance(self, instance):
        self.scope[self.key] = instance


def provider_factory(factory=_sentinel, scope=NoneScope):
    '''
    Decorator to create a provider using the given factory, and scope.
    Can also be used in a non-decorator manner.

    :param scope: Scope key, factory, or instance
    :type scope: object or callable
    :return: decorator
    :rtype: decorator
    '''
    if factory is _sentinel:
        return functools.partial(provider_factory, scope=scope)
    provider = Provider(factory, scope)
    return provider



import collections
import sys

import six

IS_PYPY = '__pypy__' in sys.builtin_module_names


def _get_object_init():
    if six.PY3 or IS_PYPY:
        return six.get_unbound_function(object.__init__)


OBJECT_INIT = _get_object_init()


class classproperty(object):

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class ProxyMutableMapping(collections.MutableMapping):

    def __init__(self, mapping):
        self.__mapping = mapping

    _fancy_repr = True

    def __repr__(self):
        if self._fancy_repr:
            return '<%s %s>' % (self.__class__.__name__, self.__mapping)
        else:
            return '%s' % dict(self)

    def __contains__(self, item):
        return item in self.__mapping

    def __getitem__(self, item):
        return self.__mapping[item]

    def __setitem__(self, key, value):
        self.__mapping[key] = value

    def __delitem__(self, key):
        del self.__mapping[key]

    def __iter__(self):
        return iter(self.__mapping)

    def __len__(self):
        return len(self.__mapping)



import mock

import test_di


class TestDi(test_di.TestDi):
    def test_inject_annotations(self, di, dependency_kv):
        key, deps = dependency_kv

        providers = dict(
            mock_provider_uno=mock.MagicMock(return_value=object()), )
        di.providers.update(providers)

        funcs = []

        def _uno_func(fn):
            funcs.append(fn)
            return fn

        @_uno_func
        @di.auto_inject()
        def _ai_inject_annotation(arg1: 'mock_provider_uno'):
            return arg1

        @_uno_func
        @di.auto_inject()
        def _ai_inject_annotation_with_return(
                arg1: 'mock_provider_uno') -> object:
            return arg1

        @_uno_func
        @di.auto_inject()
        def _ai_with_annotation(mock_provider_uno: 'an_annotation'):
            return mock_provider_uno

        @_uno_func
        @di.auto_inject('mock_provider_uno')
        def _ai_arglimited_inject_annotation(arg1: 'mock_provider_uno'):
            return arg1

        @_uno_func
        @di.auto_inject(arg1='mock_provider_uno')
        def _ai_kwoverride_with_annotation(arg1: 'an_annotation'):
            return arg1

        mpu = providers['mock_provider_uno']
        for f in funcs:
            value = f()
            assert value == mpu.return_value



"""
Note that the docs and the README double as tests themselves.
"""

import mock
import pytest
import itertools

import mainline


class TestDi(object):
    # Set of all possible scope values
    all_scopeish = set(itertools.chain(*mainline.Di.scopes.items()))

    @pytest.fixture()
    def di(self):
        di = mainline.Di()
        return di

    @pytest.fixture(params=['mock_provider0', 'mock_provider1'])
    def provider_kv(self, di, request):
        key = request.param
        provider = mock.MagicMock(return_value=object())
        di.providers[key] = provider

        def fin():
            del di.providers[key]

        request.addfinalizer(fin)
        return key, provider

    @pytest.fixture(params=dict(
        mock_deps0=set(['dep0', 'dep1', 'dep2']),
        mock_deps1=set(['dep0']),
    ).items())
    def dependency_kv(self, di, request):
        key, deps = request.param
        di.dependencies[key] = deps
        for dep in deps:
            di.providers[dep] = mock.MagicMock(return_value=object())

        def fin():
            del di.dependencies[key]
            for dep in deps:
                del di.providers[dep]

        request.addfinalizer(fin)
        return key, deps

    def test_assert_test_env(self, di):
        assert self.all_scopeish

    def test_set_instance(self, di, provider_kv):
        key, provider = provider_kv

        instance = mock.MagicMock()
        di.set_instance(key, instance)
        provider.set_instance.assert_called_once_with(instance)

    def test_get_provider(self, di, provider_kv):
        key, provider = provider_kv
        assert di.providers[key] is provider

    def test_get_provider_404(self, di):
        with pytest.raises(KeyError):
            di.providers['i_dont_exist']

    def test_get_deps(self, di, dependency_kv):
        key, deps = dependency_kv
        assert di.get_deps(key) == deps

    def test_get_missing_deps(self, di):
        key = 'mock_missing_deps'
        deps = ['missing_dep0', 'missing_dep1']
        di.dependencies[key] = set(deps)

        missing = di.get_missing_deps(key)
        assert set(missing) == set(deps)

    def test_iresolve(self, di, provider_kv):
        key, provider = provider_kv
        assert list(di.iresolve(key)) == [provider.return_value]

    def test_resolve(self, di, provider_kv):
        key, provider = provider_kv
        assert di.resolve(key) == provider.return_value
        provider.assert_called_with()

    def test_resolve_unresolvable(self, di):
        di.dependencies['missing_deps'] = set(['missing_dep0'])
        di.providers['missing_deps'] = mock.MagicMock()
        with pytest.raises(mainline.UnresolvableError):
            di.resolve('missing_deps')

    def test_resolve_many(self, di):
        providers = dict(
            mock_provider_uno=mock.MagicMock(return_value=object()),
            mock_provider_dos=mock.MagicMock(return_value=object()),
        )
        di.providers.update(providers)

        items = [(k, v.return_value) for k, v in providers.items()]
        assert di.resolve(*[x[0] for x in items]) == [x[1] for x in items]

    def test_resolve_deps(self, di, dependency_kv):
        key, deps = dependency_kv
        values = [di.resolve(dep) for dep in deps]
        assert set(di.resolve_deps(key)) == set(values)

    @pytest.mark.parametrize('scope', all_scopeish)
    def test_resolve_factory_for_each_scope(self, di, scope):
        key = 'test_factory_scope_%s' % scope
        factory = mock.MagicMock(return_value=object())
        di.register_factory(key, factory, scope=scope)

        instance = di.resolve(key)
        factory.assert_called_once_with()
        assert instance is factory.return_value

        factory.reset_mock()
        # Force that we give a different object
        factory.return_value = object()
        if scope in ['none', di.scopes['none']]:
            # NoneScope always returns a new instance
            second_instance = di.resolve(key)
            factory.assert_called_once_with()
            assert id(second_instance) != id(instance)
        else:
            # Test that provider returns the same object for multiple resolutions
            second_instance = di.resolve(key)
            di.resolve()
            factory.assert_not_called()
            assert id(second_instance) == id(instance)

    @pytest.mark.parametrize('deps', [('dep0',), ('dep0', 'dep1')])
    def test_depends_on(self, di, deps):

        @di.depends_on(*deps)
        def test():
            pass

        assert di.get_deps(test) == set(deps)



import inspect
import six
import abc

from mainline.utils import ProxyMutableMapping
from mainline.provider import IProvider

_sentinel = object()
_provider_mapping_factory = dict


class ProviderMapping(ProxyMutableMapping):
    """
    Mixin to provide mapping interface on providers.
    """

    _mapping_factory = _provider_mapping_factory

    def __init__(self, *args, **kwargs):
        if self.__class__._providers:
            self._providers = self.__class__._providers.copy()
        else:
            self._providers = self._mapping_factory()
        super(ProviderMapping, self).__init__(self._providers)
        self.update(dict(*args, **kwargs))

    def update(self, arg, allow_overwrite=False):
        """
        Update our providers from either an ICatalog subclass/instance or a mapping.

        If arg is an ICatalog, we update from it's ._providers attribute.

        :param arg: Di/Catalog/Mapping to update from.
        :type arg: ICatalog or collections.Mapping
        :param allow_overwrite: If True, allow overwriting existing keys
        :type allow_overwrite: bool
        """
        # If arg happens to be an ICatalog subclass
        if inspect.isclass(arg) and issubclass(arg, ICatalog) or isinstance(arg, ICatalog):
            arg = arg._providers
        if not allow_overwrite:
            for key in arg:
                if key in self._providers:
                    raise KeyError("Key %s already exists" % key)
        super(ProviderMapping, self).update(arg)


class ICatalog(object):
    """
    Inherit from this class to note that you support the ICatalog interface
    """

    _providers = None


class CatalogMeta(abc.ABCMeta):
    """
    Meta class used to populate providers from attributes of Catalog subclass declarations.
    """

    _provider_mapping_factory = _provider_mapping_factory

    def __new__(mcs, class_name, bases, attributes):
        cls = super(CatalogMeta, mcs).__new__(mcs, class_name, bases, attributes)

        # We may already have providers. If so, make a copy.
        if cls._providers:
            cls._providers = cls._providers.copy()
        else:
            cls._providers = mcs._provider_mapping_factory()

        cls._providers.update({k: v for k, v in six.iteritems(attributes) if isinstance(v, IProvider)})

        return cls


@six.add_metaclass(CatalogMeta)
class Catalog(ICatalog, ProviderMapping):
    pass



import functools
import inspect

import six
import wrapt

from mainline.exceptions import DiError
from mainline.utils import OBJECT_INIT, classproperty

try:
    from inspect import FullArgSpec as ArgSpec
    from inspect import getfullargspec as getargspec
except ImportError:
    from inspect import ArgSpec, getargspec


class Injector(object):

    def __init__(self, di):
        self.di = di

    def __call__(self, wrapped):
        raise NotImplementedError

    def decorate(self, wrapped):
        raise NotImplementedError


class CallableInjector(Injector):

    def __init__(self, di, *args, **kwargs):
        super(CallableInjector, self).__init__(di)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, wrapped):
        if isinstance(wrapped, six.class_types):
            cls = wrapped
            try:
                cls_init = six.get_unbound_function(cls.__init__)
                assert cls_init is not OBJECT_INIT
            except (AttributeError, AssertionError):
                raise DiError('Class %s has no __init__ to inject' % cls)
            cls.__init__ = self(cls_init)
            return cls

        if not any([self.args, self.kwargs]):
            self.injectables = self.di.get_deps(wrapped)
        else:
            self.injectables = []
            if self.args:
                self.injectables.extend(self.args)
            if self.kwargs:
                self.injectables.extend(self.kwargs.values())
            self.di.depends_on(*self.injectables)(wrapped)

        return self.decorate(wrapped)


class SpecInjector(CallableInjector):

    def decorate(self, wrapped):
        # Remove the number of args from the wrapped function's argspec
        spec = getargspec(wrapped)
        new_args = spec.args[len(self.args):]

        # Update argspec
        spec = ArgSpec(new_args, *spec[1:])

        @wrapt.decorator(adapter=spec)
        def decorator(wrapped, instance, args, kwargs):
            injected_args = list(self.di.iresolve(*self.args))

            if args:
                injected_args.extend(args)

            injected_kwargs = {
                k: self.di.resolve(v)
                for k, v in six.iteritems(self.kwargs) if k not in kwargs  # No need to resolve if we're overridden
            }

            if kwargs:
                injected_kwargs.update(kwargs)

            return wrapped(*injected_args, **injected_kwargs)

        return decorator(wrapped)


class NotFound(Exception):
    pass


class AutoSpecInjector(CallableInjector):

    def decorate(self, wrapped):
        spec = getargspec(wrapped)

        def decorator(*args, **kwargs):
            # TODO Might want auto to not be restrictable, hmm.
            # injectables = set(self.injectables or self.di.providers)
            injectables = self.di.providers

            # These are py3 only, so use getattr with a default
            spec_kwonlyargs = getattr(spec, 'kwonlyargs', [])
            spec_annotations = getattr(spec, 'annotations', {})

            def _find_injectable(arg):
                # Allow override of injected name via kwarg syntax injected_as_name=injectable_name
                arg = self.kwargs.pop(arg, arg)

                # If exact match, return that; it gets priority over annotations
                if arg in injectables:
                    return self.di.resolve(arg)

                # Py3 argument annotations: def test(blah: 'an_annotation')
                if arg in spec_annotations:
                    # Allow override of injected name via annotation (py3)
                    # Note: this should only be tried after the exact match
                    arg_annotation = spec_annotations[arg]
                    if arg_annotation in injectables:
                        return self.di.resolve(arg_annotation)

                # Nope, can't be found
                raise NotFound(arg)

            injected_args = []
            injected_kwargs = {}

            # Positional args
            args_cur_index = 0
            for arg in spec.args:
                try:
                    obj = _find_injectable(arg)
                except NotFound:
                    try:
                        obj = args[args_cur_index]
                        args_cur_index += 1
                    except IndexError:
                        # This means there aren't enough args given, which means this will most likely result in
                        # TypeError. Just let that happen so we don't hide the true exception.
                        break  # this is wanted since these are positional
                injected_args.append(obj)

            # Top it off with any remaining positional args
            remaining_args = args[args_cur_index:]
            if remaining_args:
                injected_args.extend(remaining_args)

            # Py3 keyword only args: def test(*, arg1)
            for arg in spec_kwonlyargs:
                try:
                    obj = _find_injectable(arg)
                except NotFound:
                    try:
                        obj = kwargs.pop(arg)
                    except KeyError:
                        # This means there aren't enough args given, which means this will most likely result in
                        # TypeError. Just let that happen so we don't hide the true exception.
                        continue
                injected_kwargs[arg] = obj

            injected_kwargs.update(
                {
                    k: self.di.resolve(v)
                    for k, v in six.iteritems(self.kwargs) if k not in kwargs  # No need to resolve if we're overridden
                }
            )

            if kwargs:
                injected_kwargs.update(kwargs)

            return wrapped(*injected_args, **injected_kwargs)

        return decorator


class ClassPropertyInjector(Injector):

    def __init__(self, di, key, name=None, replace_on_access=False):
        super(ClassPropertyInjector, self).__init__(di)
        self.key = key
        self.name = name
        self.replace_on_access = replace_on_access

    def _wrap_classproperty(self, cls, key, name, replace_on_access, owner=None):
        # owner is set to the instance if applicable
        val = self.di.resolve(key)
        if replace_on_access:
            setattr(cls, name, val)
        return val

    def __call__(self, klass):
        name = self.name or self.key

        # Register as dependency for klass
        self.di.depends_on(self.key)(klass)

        # Add in arguments
        partial = functools.partial(self._wrap_classproperty, klass, self.key, name, self.replace_on_access)
        # Create classproperty from it
        clsprop = classproperty(partial)

        # Attach descriptor to object
        setattr(klass, name, clsprop)

        # Return class as this is a decorator
        return klass



"""
mainline: simple python dependency injection.
"""
import pkg_resources
__version__ = pkg_resources.get_distribution(__name__).version

from mainline.exceptions import DiError, UnresolvableError, UnprovidableError

from mainline.di import Di
from mainline.catalog import Catalog
from mainline.provider import Provider, provider_factory
from mainline.scope import NoneScope, GlobalScope, ProcessScope, ThreadScope, \
    ProxyScope, NamespacedProxyScope



from __future__ import absolute_import, division, print_function, unicode_literals

import functools

from mainline.catalog import ICatalog, Catalog
from mainline.exceptions import UnresolvableError
from mainline.injection import ClassPropertyInjector, AutoSpecInjector, SpecInjector
from mainline.scope import ScopeRegistry, NoneScope, GlobalScope
from mainline.provider import Provider, provider_factory

_sentinel = object()


class Di(ICatalog):
    '''
    Dependency injection container.
    '''
    scopes = ScopeRegistry()

    # Hook these in for usability
    provider = staticmethod(provider_factory)
    Catalog = Catalog
    Provider = Provider

    # alias
    p = provider

    def __init__(self, providers_factory=Catalog, dependencies_factory=dict):
        self._providers = providers_factory()
        self._dependencies = dependencies_factory()
        super(Di, self).__init__()

    @property
    def providers(self):
        '''
        Public attribute for provider mapping
        '''
        return self._providers

    @property
    def dependencies(self):
        '''
        Public attributes for dependency mapping
        '''
        return self._dependencies

    def update(self, catalog=None, dependencies=None, allow_overwrite=False):
        '''
        Convenience method to update this Di instance with the specified contents.

        :param catalog: ICatalog supporting class or mapping
        :type catalog: ICatalog or collections.Mapping
        :param dependencies: Mapping of dependencies
        :type dependencies: collections.Mapping
        :param allow_overwrite: If True, allow overwriting existing keys. Only applies to providers.
        :type allow_overwrite: bool
        '''
        if catalog:
            self._providers.update(catalog, allow_overwrite=allow_overwrite)
        if dependencies:
            self._dependencies.update(dependencies)

    def get_deps(self, obj):
        '''
        Returns dependencies for key.

        :param key: Dependent holder key
        :type key: object
        :return: Dependencies
        :rtype: set
        '''
        return self._dependencies.get(obj, set())

    def get_missing_deps(self, obj):
        '''
        Returns missing dependencies for provider key.
        Missing meaning no instance can be provided at this time.

        :param key: Provider key
        :type key: object
        :return: Missing dependencies
        :rtype: list
        '''
        deps = self.get_deps(obj)
        ret = []
        for key in deps:
            provider = self._providers.get(key)
            if provider and provider.providable:
                continue
            ret.append(key)
        return ret

    def iresolve(self, *keys):
        '''
        Iterates over resolved instances for given provider keys.

        :param keys: Provider keys
        :type keys: tuple
        :return: Iterator of resolved instances
        :rtype: generator
        '''
        for key in keys:
            missing = self.get_missing_deps(key)
            if missing:
                raise UnresolvableError("Missing dependencies for %s: %s" % (key, missing))

            provider = self._providers.get(key)
            if not provider:
                raise UnresolvableError("Provider does not exist for %s" % key)

            yield provider()

    def resolve(self, *keys):
        '''
        Returns resolved instances for given provider keys.

        If only one positional argument is given, only one is returned.

        :param keys: Provider keys
        :type keys: tuple
        :return: Resolved instance(s); if only one key given, otherwise list of them.
        :rtype: object or list
        '''
        instances = list(self.iresolve(*keys))
        if len(keys) == 1:
            return instances[0]
        return instances

    def resolve_deps(self, obj):
        '''
        Returns list of resolved dependencies for given obj.

        :param obj: Object to lookup dependencies for
        :type obj: object
        :return: Resolved dependencies
        :rtype: list
        '''
        deps = self.get_deps(obj)
        return list(self.iresolve(*deps))

    def register_factory(self, key, factory=_sentinel, scope=NoneScope, allow_overwrite=False):
        '''
        Creates and registers a provider using the given key, factory, and scope.
        Can also be used as a decorator.

        :param key: Provider key
        :type key:  object
        :param factory: Factory callable
        :type factory: callable
        :param scope: Scope key, factory, or instance
        :type scope: object or callable
        :return: Factory (or None if we're creating a provider without a factory)
        :rtype: callable or None
        '''
        if factory is _sentinel:
            return functools.partial(self.register_factory, key, scope=scope)
        if not allow_overwrite and key in self._providers:
            raise KeyError("Key %s already exists" % key)
        provider = self.provider(factory, scope)
        self._providers[key] = provider
        return factory

    f = register_factory

    def set_instance(self, key, instance, default_scope=GlobalScope):
        '''
        Sets instance under specified provider key. If a provider for specified key does not exist, one is created
        without a factory using the given scope.

        :param key: Provider key
        :type key: object
        :param instance: Instance
        :type instance: object
        :param default_scope: Scope key, factory, or instance
        :type default_scope: object or callable
        '''
        if key not in self._providers:
            # We don't know how to create this kind of instance at this time, so add it without a factory.
            factory = None
            self.register_factory(key, factory, default_scope)
        self._providers[key].set_instance(instance)

    def depends_on(self, *keys):
        '''
        Decorator that marks the wrapped as depending on specified provider keys.

        :param keys: Provider keys to mark as dependencies for wrapped
        :type keys: tuple
        :return: decorator
        :rtype: decorator
        '''

        def decorator(wrapped):
            if keys:
                if wrapped not in self._dependencies:
                    self._dependencies[wrapped] = set()
                self._dependencies[wrapped].update(keys)
            return wrapped

        return decorator

    def inject_classproperty(self, key, name=None, replace_on_access=False):
        '''
        Decorator that injects the specified key as a classproperty.

        If replace_on_access is True, then it replaces itself with the instance on first lookup.

        :param key: Provider key
        :type key: object
        :param name: Name of classproperty, defaults to key
        :type name: str
        :param replace_on_access: If True, replace the classproperty with the actual value on first lookup
        :type replace_on_access: bool
        '''
        return ClassPropertyInjector(self, key, name=name, replace_on_access=replace_on_access)

    def inject_exact(self, *args, **kwargs):
        '''
        Decorator that injects the specified arguments when the wrapped is called. Argspec is modified on the wrapped
        accordingly.

        Positional arguments are injected in the order they are given.

        Keyword arguments are injected with the given key name, eg omg='whoa' would inject as omg=<whoa instance>.

        Any specified provider args are added as dependencies for the wrapped.

        :param args: Provider args to inject as positional arguments
        :type args: tuple
        :param kwargs: Mapping of keyword argument name to provider args to inject as keyword arguments
        :type kwargs: dict
        :return: decorator
        :rtype: decorator
        '''
        return SpecInjector(self, *args, **kwargs)

    def auto_inject(self, *args, **kwargs):
        '''
        Decorator that magically inspects the argspec of the wrapped upon call, injecting provider instances as names
        match.

        Performance wise, this is currently slower than `inject_exact` as more is done each runtime to determine what
        needs to be injected.

        Positional arguments are added as dependencies to wrapped.

        Keyword arguments are handled similarly to inject(), being a mapping of keyword argument name to provider key.

        Any specified provider args are added as dependencies for the wrapped.

        If dependencies are associated with the wrapped, only those are checked for match in the argspec.

        :param args: Provider args to inject as positional arguments
        :type args: tuple
        :param kwargs: Mapping of keyword argument name to provider args to inject as keyword arguments
        :type kwargs: dict
        :return: decorator
        :rtype: decorator
        '''
        return AutoSpecInjector(self, *args, **kwargs)

    inject = inject_exact

    # This makes us callable, default to inject_exact for now.
    __call__ = inject

    # aliases
    i = inject
    ai = auto_inject




#!/usr/bin/env python

from setuptools import setup
import sys

conf = dict(
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    pbr=True,
)

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
if needs_pytest:
    conf['setup_requires'].append('pytest-runner')

setup(**conf)



import os
import threading
import collections

from mainline.utils import ProxyMutableMapping


class IScope(ProxyMutableMapping):
    register = False
    name = None

    instances = None
    instances_factory = dict

    def __init__(self, *args, **kwargs):
        if self.instances is None:
            self.instances = self.instances_factory()
        super(IScope, self).__init__(self.instances)
        self.update(dict(*args, **kwargs))

    def __str__(self):
        return self.name

    def __key_transform__(self, key):
        return key

    class _key(str):
        pass

    def _key_factory(self, key):
        if not isinstance(key, self._key):
            key = self.__key_transform__(key)
            key = self._key(key)
        return key

    def __contains__(self, key):
        key = self._key_factory(key)
        return super(IScope, self).__contains__(key)

    def __getitem__(self, key):
        key = self._key_factory(key)
        return super(IScope, self).__getitem__(key)

    def __setitem__(self, key, value):
        key = self._key_factory(key)
        super(IScope, self).__setitem__(key, value)

    def __delitem__(self, key):
        key = self._key_factory(key)
        super(IScope, self).__delitem__(key)


class NoneScope(IScope):
    register = True
    name = 'none'

    def __setitem__(self, key, value):
        return


class GlobalScope(IScope):
    register = True
    name = 'global'


class SingletonScope(GlobalScope):
    """ Alias for GlobalScope
    """


class ProcessScope(IScope):
    register = True
    name = 'process'

    def __key_transform__(self, key):
        return '%s_%s' % (os.getpid(), key)


class ThreadScope(IScope):
    register = True
    name = 'thread'

    def __init__(self, *args, **kwargs):
        self._thread_local = threading.local()
        super(ThreadScope, self).__init__(*args, **kwargs)

    def instances_factory(self):
        if not hasattr(self._thread_local, 'instances'):
            self._thread_local.instances = dict()
        return self._thread_local.instances


class ProxyScope(IScope):

    def __init__(self, scope, *args, **kwargs):
        self.instances = scope
        super(ProxyScope, self).__init__(*args, **kwargs)


class NamespacedProxyScope(ProxyScope):

    def __init__(self, namespace, scope, *args, **kwargs):
        self.namespace = namespace
        super(NamespacedProxyScope, self).__init__(scope, *args, **kwargs)

    @property
    def name(self):
        return self.namespace

    def __key_transform__(self, key):
        return '%s__%s' % (self.namespace, key)


SCOPE_FACTORIES = {}


class ScopeRegistry(ProxyMutableMapping):
    _factories = SCOPE_FACTORIES

    def __init__(self):
        super(ScopeRegistry, self).__init__(self._factories)
        self._build()

    def _build(self):
        classes = IScope.__subclasses__()
        classes = filter(lambda x: x.register, classes)
        list(map(self.register_factory, classes))

    def register_factory(self, factory, name=None):
        if name is None:
            # name = str(factory)
            name = getattr(factory, 'name', None)
        if name:
            self._factories[name] = factory
        self._factories[factory] = factory

    def resolve(self, scope_or_scope_factory, instantiate_factory=True):
        if self.is_scope_instance(scope_or_scope_factory):
            instance = scope_or_scope_factory
            return instance

        elif self.is_scope_factory(scope_or_scope_factory):
            factory = scope_or_scope_factory
            if not instantiate_factory:
                return factory
            instance = factory()
            return instance

        elif scope_or_scope_factory in self._factories:
            factory = self._factories[scope_or_scope_factory]
            return self.resolve(factory)

        else:
            raise KeyError("Scope %s is not known" % scope_or_scope_factory)

    _scope_type = collections.MutableMapping

    @classmethod
    def is_scope_factory(cls, obj):
        return callable(obj) and issubclass(obj, cls._scope_type)

    @classmethod
    def is_scope_instance(cls, obj):
        return isinstance(obj, cls._scope_type)

    @classmethod
    def is_scope(cls, obj):
        return cls.is_scope_factory(obj) or cls.is_scope_instance(obj)
